@@ -73,6 +73,7 @@ from hypothesis.internal.floats import (
 from hypothesis.internal.reflection import (
     define_function_signature,
     deprecated_posargs,
+    get_pretty_function_description,
     is_typed_named_tuple,
     nicerepr,
     proxies,
@@ -118,7 +119,6 @@ from hypothesis.strategies._internal.strategies import (
 from hypothesis.strategies._internal.strings import (
     FixedSizeBytes,
     OneCharStringStrategy,
-    StringStrategy,
 )
 from hypothesis.utils.conventions import InferType, infer, not_set
 
@@ -191,12 +191,18 @@ def cacheable(fn: T) -> T:
     return cached_strategy
 
 
-def base_defines_strategy(force_reusable: bool) -> Callable[[T], T]:
+def defines_strategy(
+    *, force_reusable_values: bool = False, try_non_lazy: bool = False
+) -> Callable[[T], T]:
     """Returns a decorator for strategy functions.
 
     If force_reusable is True, the generated values are assumed to be
     reusable, i.e. immutable and safe to cache, across multiple test
     invocations.
+
+    If try_non_lazy is True, attempt to execute the strategy definition
+    function immediately, so that a LazyStrategy is only returned if this
+    raises an exception.
     """
 
     def decorator(strategy_definition):
@@ -206,8 +212,18 @@ def base_defines_strategy(force_reusable: bool) -> Callable[[T], T]:
 
         @proxies(strategy_definition)
         def accept(*args, **kwargs):
+            if try_non_lazy:
+                # Why not try this unconditionally?  Because we'd end up with very
+                # deep nesting of recursive strategies - better to be lazy unless we
+                # *know* that eager evaluation is the right choice.
+                try:
+                    return strategy_definition(*args, **kwargs)
+                except Exception:
+                    # If invoking the strategy definition raises an exception,
+                    # wrap that up in a LazyStrategy so it happens again later.
+                    pass
             result = LazyStrategy(strategy_definition, args, kwargs)
-            if force_reusable:
+            if force_reusable_values:
                 result.force_has_reusable_values = True
                 assert result.has_reusable_values
             return result
@@ -218,10 +234,6 @@ def base_defines_strategy(force_reusable: bool) -> Callable[[T], T]:
     return decorator
 
 
-defines_strategy = base_defines_strategy(False)
-defines_strategy_with_reusable_values = base_defines_strategy(True)
-
-
 class Nothing(SearchStrategy):
     def calc_is_empty(self, recur):
         return True
@@ -274,7 +286,7 @@ def just(value: T) -> SearchStrategy[T]:
     return JustStrategy(value)
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 def none() -> SearchStrategy[None]:
     """Return a strategy which only generates None.
 
@@ -331,7 +343,7 @@ def one_of(*args):  # noqa: F811
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 def integers(min_value: int = None, max_value: int = None) -> SearchStrategy[int]:
     """Returns a strategy which generates integers.
 
@@ -388,18 +400,18 @@ def integers(min_value: int = None, max_value: int = None) -> SearchStrategy[int
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 def booleans() -> SearchStrategy[bool]:
     """Returns a strategy which generates instances of :class:`python:bool`.
 
     Examples from this strategy will shrink towards ``False`` (i.e.
     shrinking will replace ``True`` with ``False`` where possible).
     """
-    return sampled_from([False, True])
+    return SampledFromStrategy([False, True], repr_="booleans()")
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def floats(
     min_value: Real = None,
@@ -617,7 +629,7 @@ def floats(
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 def tuples(*args: SearchStrategy) -> SearchStrategy[tuple]:
     """Return a strategy which generates a tuple of the same length as args by
     generating the value at index i from args[i].
@@ -644,7 +656,7 @@ def sampled_from(elements: Type[enum.Enum]) -> SearchStrategy[Any]:
     pass  # pragma: no cover
 
 
-@defines_strategy  # noqa: F811
+@defines_strategy(try_non_lazy=True)  # noqa: F811
 def sampled_from(elements):
     """Returns a strategy which generates any value present in ``elements``.
 
@@ -665,17 +677,23 @@ def sampled_from(elements):
         raise InvalidArgument("Cannot sample from a length-zero sequence.")
     if len(values) == 1:
         return just(values[0])
+    if isinstance(elements, type) and issubclass(elements, enum.Enum):
+        repr_ = "sampled_from(%s.%s)" % (elements.__module__, elements.__name__)
+    else:
+        repr_ = "sampled_from(%r)" % (elements,)
     if hasattr(enum, "Flag") and isclass(elements) and issubclass(elements, enum.Flag):
         # Combinations of enum.Flag members are also members.  We generate
         # these dynamically, because static allocation takes O(2^n) memory.
-        return sets(sampled_from(values), min_size=1).map(
+        # LazyStrategy is used for the ease of force_repr.
+        inner = sets(sampled_from(list(values)), min_size=1).map(
             lambda s: reduce(operator.or_, s)
         )
-    return SampledFromStrategy(elements)
+        return LazyStrategy(lambda: inner, args=[], kwargs={}, force_repr=repr_)
+    return SampledFromStrategy(values, repr_)
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def lists(
     elements: SearchStrategy[Ex],
@@ -766,7 +784,7 @@ def lists(
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def sets(
     elements: SearchStrategy[Ex], *, min_size: int = 0, max_size: int = None
@@ -786,7 +804,7 @@ def sets(
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def frozensets(
     elements: SearchStrategy[Ex], *, min_size: int = 0, max_size: int = None
@@ -813,7 +831,7 @@ class PrettyIter:
         return "iter({!r})".format(self._values)
 
 
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def iterables(
     elements: SearchStrategy[Ex],
@@ -839,22 +857,25 @@ def iterables(
     ).map(PrettyIter)
 
 
-@defines_strategy
+@defines_strategy()
 def fixed_dictionaries(
     mapping: Dict[T, SearchStrategy[Ex]],
     *,
     optional: Dict[T, SearchStrategy[Ex]] = None
 ) -> SearchStrategy[Dict[T, Ex]]:
     """Generates a dictionary of the same type as mapping with a fixed set of
-    keys mapping to strategies. mapping must be a dict subclass.
+    keys mapping to strategies. ``mapping`` must be a dict subclass.
 
     Generated values have all keys present in mapping, with the
     corresponding values drawn from mapping[key]. If mapping is an
     instance of OrderedDict the keys will also be in the same order,
     otherwise the order is arbitrary.
 
+    If ``optional`` is passed, the generated value *may or may not* contain each
+    key from ``optional`` and a value drawn from the corresponding strategy.
+
     Examples from this strategy shrink by shrinking each individual value in
-    the generated dictionary.
+    the generated dictionary, and omitting optional key-value pairs.
     """
     check_type(dict, mapping, "mapping")
     for k, v in mapping.items():
@@ -878,7 +899,7 @@ def fixed_dictionaries(
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def dictionaries(
     keys: SearchStrategy[Ex],
@@ -914,7 +935,7 @@ def dictionaries(
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def characters(
     *,
@@ -1021,7 +1042,7 @@ def characters(
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def text(
     alphabet: Union[Sequence[str], SearchStrategy[str]] = characters(
@@ -1073,11 +1094,11 @@ def text(
         )
     if (max_size == 0 or char_strategy.is_empty) and not min_size:
         return just("")
-    return StringStrategy(lists(char_strategy, min_size=min_size, max_size=max_size))
+    return lists(char_strategy, min_size=min_size, max_size=max_size).map("".join)
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def from_regex(
     regex: Union[AnyStr, Pattern[AnyStr]], *, fullmatch: bool = False
@@ -1118,7 +1139,7 @@ def from_regex(
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def binary(*, min_size: int = 0, max_size: int = None) -> SearchStrategy[bytes]:
     """Generates :class:`python:bytes`.
@@ -1138,7 +1159,7 @@ def binary(*, min_size: int = 0, max_size: int = None) -> SearchStrategy[bytes]:
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 def randoms(
     *, note_method_calls: bool = False, use_true_random: bool = None
 ) -> SearchStrategy[random.Random]:
@@ -1191,7 +1212,7 @@ class RandomModule(SearchStrategy):
 
 
 @cacheable
-@defines_strategy
+@defines_strategy()
 def random_module() -> SearchStrategy[RandomSeeder]:
     """The Hypothesis engine handles PRNG state for the stdlib and Numpy random
     modules internally, always seeding them to zero and restoring the previous
@@ -1214,7 +1235,7 @@ def random_module() -> SearchStrategy[RandomSeeder]:
 # Note that for the benefit of documentation and introspection tools, we set the
 # __signature__ attribute to show the semantic rather than actual signature.
 @cacheable
-@defines_strategy
+@defines_strategy()
 def builds(
     *callable_and_args: Union[Callable[..., Ex], SearchStrategy[Any]],
     **kwargs: Union[SearchStrategy[Any], InferType]
@@ -1442,6 +1463,11 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
     # convert empty results into an explicit error.
     if thing in types._global_type_lookup:
         return as_strategy(types._global_type_lookup[thing], thing)
+    # We also have a special case for TypeVars.
+    # They are represented as instances like `~T` when they come here.
+    # We need to work with their type instead.
+    if isinstance(thing, TypeVar) and type(thing) in types._global_type_lookup:  # type: ignore
+        return as_strategy(types._global_type_lookup[type(thing)], thing)
     # If there's no explicitly registered strategy, maybe a subtype of thing
     # is registered - if so, we can resolve it to the subclass strategy.
     # We'll start by checking if thing is from from the typing module,
@@ -1493,7 +1519,7 @@ def _from_type(thing: Type[Ex]) -> SearchStrategy[Ex]:
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def fractions(
     min_value: Union[Real, str] = None,
@@ -1603,7 +1629,7 @@ def _as_finite_decimal(
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def decimals(
     min_value: Union[Real, str] = None,
@@ -1739,7 +1765,7 @@ class PermutationStrategy(SearchStrategy):
         return result
 
 
-@defines_strategy
+@defines_strategy()
 def permutations(values: Sequence[T]) -> SearchStrategy[List[T]]:
     """Return a strategy which returns permutations of the ordered collection
     ``values``.
@@ -1796,7 +1822,7 @@ def composite(f: Callable[..., Ex]) -> Callable[..., SearchStrategy[Ex]]:
     }
     new_argspec = argspec._replace(args=argspec.args[1:], annotations=annots)
 
-    @defines_strategy
+    @defines_strategy()
     @define_function_signature(f.__name__, f.__doc__, new_argspec)
     def accept(*args, **kwargs):
         return CompositeStrategy(f, args, kwargs)
@@ -1805,7 +1831,7 @@ def composite(f: Callable[..., Ex]) -> Callable[..., SearchStrategy[Ex]]:
     return accept
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @cacheable
 @deprecated_posargs
 def complex_numbers(
@@ -1915,7 +1941,7 @@ def shared(base: SearchStrategy[Ex], *, key: Hashable = None) -> SearchStrategy[
 
 
 @cacheable
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def uuids(*, version: int = None) -> SearchStrategy[UUID]:
     """Returns a strategy that generates :class:`UUIDs <uuid.UUID>`.
@@ -1960,7 +1986,7 @@ class RunnerStrategy(SearchStrategy):
             return runner
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 @deprecated_posargs
 def runner(*, default: Any = not_set) -> SearchStrategy[Any]:
     """A strategy for getting "the current test runner", whatever that may be.
@@ -2062,6 +2088,12 @@ def register_type_strategy(
 
     ``strategy`` may be a search strategy, or a function that takes a type and
     returns a strategy (useful for generic types).
+
+    Note that you may not register a parametrised generic type (such as
+    ``MyCollection[int]``) directly, because the resolution logic does not
+    handle this case correctly.  Instead, you may register a *function* for
+    ``MyCollection`` and `inspect the type parameters within that function
+    <https://stackoverflow.com/q/48572831>`__.
     """
     # TODO: We would like to move this to the top level, but pending some major
     # refactoring it's hard to do without creating circular imports.
@@ -2076,6 +2108,28 @@ def register_type_strategy(
         )
     elif isinstance(strategy, SearchStrategy) and strategy.is_empty:
         raise InvalidArgument("strategy=%r must not be empty")
+    elif types.has_type_arguments(custom_type):
+        origin = getattr(custom_type, "__origin__", None)
+        if callable(strategy):
+            strategy_repr = get_pretty_function_description(strategy)
+        else:
+            strategy_repr = repr(strategy)
+        note_deprecation(
+            "Registering a generic type with arguments (%r) is deprecated, and "
+            "will be an error in future, because the resolution logic is badly "
+            "broken.  Instead, register a function for the origin type (%r) "
+            "which can inspect specific type objects and return a strategy.  "
+            "%s will be registered for any type arguments."
+            % (custom_type, origin, strategy_repr),
+            since="2020-08-17",
+        )
+        if origin in types._global_type_lookup:
+            raise InvalidArgument(
+                "Cannot register %r, because the without-arguments form %r already "
+                "has a registered strategy %s" % (custom_type, origin, strategy_repr)
+            )
+        custom_type = origin
+
     types._global_type_lookup[custom_type] = strategy
     from_type.__clear_cache()  # type: ignore
 
@@ -2114,7 +2168,7 @@ def deferred(definition: Callable[[], SearchStrategy[Ex]]) -> SearchStrategy[Ex]
     return DeferredStrategy(definition)
 
 
-@defines_strategy_with_reusable_values
+@defines_strategy(force_reusable_values=True)
 def emails() -> SearchStrategy[str]:
     """A strategy for generating email addresses as unicode strings. The
     address format is specified in :rfc:`5322#section-3.4.1`. Values shrink
@@ -2133,10 +2187,13 @@ def emails() -> SearchStrategy[str]:
     )
 
 
-@defines_strategy
+@defines_strategy()
 @deprecated_posargs
 def functions(
-    *, like: Callable[..., Any] = lambda: None, returns: SearchStrategy[Any] = None
+    *,
+    like: Callable[..., Any] = lambda: None,
+    returns: SearchStrategy[Any] = None,
+    pure: bool = False
 ) -> SearchStrategy[Callable[..., Any]]:
     # The proper type signature of `functions()` would have T instead of Any, but mypy
     # disallows default args for generics: https://github.com/python/mypy/issues/3737
@@ -2149,12 +2206,17 @@ def functions(
     for the function is drawn from the ``returns`` argument, which must be a
     strategy.
 
-    Note that the generated functions do not validate their arguments, and
+    If ``pure=True``, all arguments passed to the generated function must be
+    hashable, and if passed identical arguments the original return value will
+    be returned again - *not* regenerated, so beware mutable values.
+
+    If ``pure=False``, generated functions do not validate their arguments, and
     may return a different value if called again with the same arguments.
 
     Generated functions can only be called within the scope of the ``@given``
     which created them.  This strategy does not support ``.example()``.
     """
+    check_type(bool, pure, "pure")
     if not callable(like):
         raise InvalidArgument(
             "The first argument to functions() must be a callable to imitate, "
@@ -2166,7 +2228,7 @@ def functions(
         returns = from_type(hints.get("return", type(None)))
 
     check_strategy(returns, "returns")
-    return FunctionStrategy(like, returns)
+    return FunctionStrategy(like, returns, pure)
 
 
 @composite
